'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var jsep = require('../../parser/jsep.cjs');
var builders = require('../ast/builders.cjs');
var dump = require('../dump.cjs');
var fnParams = require('../templates/fn-params.cjs');
var internalScope = require('../templates/internal-scope.cjs');
var scope = require('../templates/scope.cjs');
var treeMethodCall = require('../templates/tree-method-call.cjs');
var commonjs = require('./modules/commonjs.cjs');
var esm = require('./modules/esm.cjs');
var traversalZones = require('./traversal-zones.cjs');

const params = [builders.identifier('input'), builders.identifier('callbacks')];

const NEW_SCOPE_VARIABLE_DECLARATION = builders.variableDeclaration('const', [
  builders.variableDeclarator(scope.default._, builders.newExpression(builders.identifier('Scope'), params)),
]);

/*
import {
 // deps
} from 'nimma/runtime';
// placement: tree
const tree = {};

// placement: program

export default function (input, callbacks) {
  const scope = new Scope(input, callbacks);

  try {
    // placement: body

    scope.traverse(() => {
      // placement: traverse
    });
  } finally {
    scope.destroy();
  }
}
*/

class ESTree {
  #tree = builders.objectExpression([]);
  #shorthands = builders.objectExpression([]);
  #runtimeDependencies;
  #program = new Set();
  #body = new Set();
  #traverse = new Set();
  #availableShorthands;
  #states = -1;

  constructor({ customShorthands, module }) {
    this.module = module;
    this.ctx = null;
    this.traversalZones = new traversalZones.default();
    this.#availableShorthands = customShorthands;
    this.#runtimeDependencies = new Map([['Scope', 'Scope']]);
  }

  addRuntimeDependency(specifier) {
    this.#runtimeDependencies.set(specifier, specifier);
  }

  attachCustomShorthand(name) {
    if (
      this.#availableShorthands === null ||
      !(name in this.#availableShorthands)
    ) {
      throw new ReferenceError(`Shorthand '${name}' is not defined`);
    }

    this.#shorthands.properties.push(
      builders.objectMethod(
        'method',
        builders.identifier(name),
        fnParams.statefulFnParams,
        builders.blockStatement([
          builders.returnStatement(jsep.default(this.#availableShorthands[name])),
        ]),
      ),
    );
  }

  getMethodByHash(hash) {
    return this.#tree.properties.find(prop => prop.key.value === hash);
  }

  push(node, placement) {
    switch (placement) {
      case 'tree-method':
      case 'stateful-tree-method':
        this.#tree.properties.push(
          builders.objectMethod(
            'method',
            builders.stringLiteral(this.ctx.id),
            placement === 'stateful-tree-method'
              ? fnParams.statefulFnParams
              : fnParams.statelessFnParams,
            node,
          ),
        );
        break;
      case 'program':
        if (!this.#program.has(node)) {
          this.#program.add(node);
        }

        break;
      case 'body':
        if (!this.#body.has(node)) {
          this.#body.add(node);
        }

        break;
      case 'traverse':
        this.#traverse.add(treeMethodCall.default(node.value, fnParams.statelessFnParams));
        break;
      case 'stateful-traverse': {
        this.#states += 1;
        const id = builders.identifier(`state${this.#states}`);
        this.push(
          builders.variableDeclaration('const', [
            builders.variableDeclarator(id, builders.callExpression(scope.default.allocState, [])),
          ]),
          'body',
        );

        this.#traverse.add(treeMethodCall.default(node.value, [scope.default._, id]));
        break;
      }
    }
  }

  pushAll(items) {
    for (const item of items) {
      this.push(...item);
    }
  }

  export(format) {
    const traversalZones = this.traversalZones.root;

    const program = builders.program(
      [
        ...this.#program,
        traversalZones,
        this.#tree.properties.length === 0
          ? null
          : builders.variableDeclaration('const', [
              builders.variableDeclarator(internalScope.default.tree, this.#tree),
            ]),
        this.#shorthands.properties.length === 0
          ? null
          : builders.variableDeclaration('const', [
              builders.variableDeclarator(internalScope.default.shorthands, this.#shorthands),
            ]),
        builders.functionDeclaration(
          null,
          params,
          builders.blockStatement(
            [
              NEW_SCOPE_VARIABLE_DECLARATION,
              builders.tryStatement(
                builders.blockStatement(
                  [
                    ...this.#body,
                    this.#traverse.size === 0
                      ? null
                      : builders.expressionStatement(
                          builders.callExpression(scope.default.traverse, [
                            builders.arrowFunctionExpression(
                              [],
                              builders.blockStatement(Array.from(this.#traverse)),
                            ),
                            traversalZones === null
                              ? builders.nullLiteral()
                              : traversalZones.declarations[0].id,
                          ]),
                        ),
                  ].filter(Boolean),
                ),
                null,
                builders.blockStatement([
                  builders.expressionStatement(builders.callExpression(scope.default.destroy, [])),
                ]),
              ),
            ].filter(Boolean),
          ),
        ),
      ].filter(Boolean),
    );

    const mod = format === 'esm' ? esm.default : commonjs.default;
    mod(Array.from(this.#runtimeDependencies), program);

    return dump.default(program);
  }
}

exports.default = ESTree;
